#!/usr/bin/env python
# ======================================================================
# Copyright (C) 2008 and onwards Robert Muth <robert at muth dot org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ======================================================================

"""module for converting between different audio formats"""
import logging
import os
import urllib
import fileinfo
import Queue
import shutil
import threading
import time

__all__ = ["AddPrefetch",
           "Init",
           "SubmitJob",
           ]

# ======================================================================
# Note: ubuntu packages providing these tools
# lame: lame
# flac: flac
# oggenc: vorbis-tools
# mpc123: /usr/bin/mpc123


ENCODE_COMMANDS = {
    ".mp3":
    ('lame '
     '--replaygain-fast --preset fast standard '
     '--id3v2-only '
     '--tt "%(title)s" '
     '--ta "%(artist)s" '
     '--tl "%(album)s" '
     '--tc "%(comment)s" '
     '--tn "%(track)s" '
     '--tg "%(genre)s" '
     '--ty "%(year)s" '
     '- "%(dst)s"'),

    ".ogg":
    ('oggenc '
     '-q7 '
     '--title "%(title)s" '
     '--artist "%(artist)s" '
     '--album "%(album)s" '
     '--comment "comm::xxx=%(comment)s" '
     '--tracknum "%(track)s" '
     '--genre "%(genre)s" '
     '--output "%(dst)s" '
     '-'),
}


DECODE_COMMANDS = {
    ".wav":  'cat "%(src)s"',
    ".flac": 'flac -d -c "%(src)s"',
    ".mpc":  'mppdec-static "%(src)s"',
    ".mp3":  'lame --decode "%(src)s" -',
    ".mpc":  'mpc123 --wav - "%(src)s"',
    ".ogg":  'oggdec -o - "%(src)s"',
}


# thread pool:
# http://code.activestate.com/recipes/475160/
# ======================================================================
# where to store the converted files
CACHE_DIR = None
# which other conversations to trigger if a certain conversion is requested
PREFETCH_MAP = {}
THREAD_POOL = None
JOB_HISTORY = []

def FormatTime(t):
    return time.strftime("%Y/%m/%d %a", t)
# ======================================================================
class ThreadPool(Queue.Queue):

    def __init__(self, num_workers):
        Queue.Queue.__init__(self)
        for i in range(num_workers):
            t = threading.Thread(target=self.worker)
            t.setDaemon(True)
            t.start()

    def worker(self):
        while 1:
            item = self.get()
            item.Run()
        return

# ======================================================================
class ConversionJob:

    STATUS_SUBMITTED = 0
    STATUS_IN_PROGRESS = 1
    STATUS_COMPLETE = 2
    STATUS_FAILED = 3

    def __init__(self, file_src, file_dst):
        global JOB_HISTORY
        self._start = time.time()
        self._stop = None
        self._status = 0
        self._file_src = file_src
        self._file_dst = file_dst
        JOB_HISTORY.append(self)
        return

    def IsComplete(self):
        return self._status in (ConversionJob.STATUS_COMPLETE,
                                ConversionJob.STATUS_FAILED)

    def IsFailure(self):
        return self._status == ConversionJob.STATUS_FAILED

    def SetStatus(self, result):
        self._stop = time.time()
        self._status = result
        return

    def GetCovertedFile(self):
        return self._file_dst

    def Render(self):
        time_or_status = 0
        return (FormatTime(self._start),
                time_or_status,
                self._file_src,
                self._file_dst)

    def Run(self):
        _, ext_src = os.path.splitext(self._file_src)
        _, ext_dst = os.path.splitext(self._file_dst)
        assert ext_src in DECODE_COMMANDS
        assert ext_dst in ENCODE_COMMANDS
        fi = fileinfo.FileInfo(self._file_src, 1)
        d = {
            "title": fi.title,
            "artist": fi.artist,
            "album": fi.album,
            "track": fi.track,
            "year": fi.year,
            "comment": fi.comment,
            "genre": fi.genre,
            "src": self._file_src,
            "dst": self._file_dst + ".tmp"
            }
        command = (DECODE_COMMANDS[ext_src] % d +
                   " 2>/dev/null | " +
                   ENCODE_COMMANDS[ext_dst] % d + " 2>/dev/null")
        logging.info("conversion command: [%s]", command)
        result = os.system(command)
        logging.info("conversion complete result: %d", result)
        if not os.path.exists(self._file_dst + ".tmp"):
            logging.error("conversion failed [%s] -> [%s]",
                          self._file_src, self._file_dst)
            self.SetStatus(ConversionJob.STATUS_FAILED)
            return
        os.rename(self._file_dst + ".tmp", self._file_dst)
        self.SetStatus(ConversionJob.STATUS_COMPLETE)
        logging.info("conversion success [%s] -> [%s]",
                      self._file_src, self._file_dst)
        return

    def WaitForCompletion(self, delta=1):
        # TODO: avoid polling
        while not  self.IsComplete():
            time.sleep(delta)
        return


# ======================================================================
def CreateCachePath(fullpath):
    global CACHE_DIR
    cache_path = urllib.quote(fullpath)
    cache_path = cache_path.replace("/", "%2f")
    return os.path.join(CACHE_DIR, cache_path)

# ======================================================================
def Init(cache_dir, clear, nthreads):
    """Must be called before anything else"""
    global CACHE_DIR
    global THREAD_POOL
    CACHE_DIR = cache_dir
    THREAD_POOL = ThreadPool(nthreads)
    if clear and os.path.isdir(CACHE_DIR):
        shutil.rmtree(CACHE_DIR)
    if not os.path.isdir(CACHE_DIR):
        os.mkdir(CACHE_DIR)


def AddPrefetch(file1, file2):
    global PREFETCH_MAP
    PREFETCH_MAP[file1] = file2


def SubmitJob(file_src, ext_dst, force=False):
    global THREAD_POOL
    if not os.path.exists(file_src):
        logging.error("source file does not exist: [%s]", file_src)
        job = ConversionJob(file_src, "")
        job.SetStatus(ConversionJob.STATUS_FAILED)
        return job

    base, ext_src = os.path.splitext(file_src)
    file_dst = base + ext_dst
    if not force and os.path.exists(file_dst):
        logging.warning("target exists: %s", file_dst)
        job = ConversionJob(file_src, file_dst)
        job.SetStatus(ConversionJob.STATUS_COMPLETE)
        return job

    assert file_src != file_dst
    file_cache = CreateCachePath(file_dst)
    job = ConversionJob(file_src, file_cache)

    if not force and os.path.exists(file_cache):
        logging.info("cached target exists: %s", file_cache)
        job.SetStatus(ConversionJob.STATUS_COMPLETE)
        return job

    logging.info("submitting convert %s -> %s", file_src, ext_dst)
    THREAD_POOL.put(job)
    return job


def PrefetchJob(file_src, ext_dst):
    global PREFETCH_MAP
    if file_src in PREFETCH_MAP:
        logging.info("prefetch for %s is %s", file_src, PREFETCH_MAP[file_src])
        return SubmitJob(PREFETCH_MAP[file_src], ext_dst)
    return None
