#!/usr/bin/env python
# ======================================================================
# Copyright (C) 2008 and onwards Robert Muth <robert at muth dot org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ======================================================================


__all__ = ["FileInfo",
           "CreateExtInfoOrEmpty",
           "CreateIcyNameOrEmpty",
           ]

import os
import re
import urllib

import mutagen

# a pattern used to trim leading digits, spaces, and dashes from a song
### would be nice to get a bit fancier with the possible trimming
re_trim = re.compile('[-0-9 ]*-[ ]*(.*)')

def RenderFilename(filename):
    # if a song has a prefix that matches the directory, and something
    # exists after that prefix, then strip it. don't strip if the
    # directory is a single-letter.
    # TODO: revive this
    #if len(base) > thisdirlen > 1 and base[:thisdirlen] == thisdir:
    #  base = base[thisdirlen:]

    # trim a bit of stuff off of the file
    base = os.path.basename(filename)
    base = urllib.unquote(base)
    match = re_trim.match(base)
    if match:
        base = match.group(1)
    return base

class FileInfo:
  """Grab as much info as you can from the file given"""


  def RenderFileInfo(self, filename):
      if not self.valid:
          return RenderFilename(filename)
      data = "%s - %s" % (self.artist, self.title)
      if len(data) >= 5:
          return data
      return RenderFilename(filename)


  def GetDict(self):
      meta = {}
      meta["artist"] = self.artist
      meta["title"] = self.title
      meta["duration"] = str(int(self.total_time))
      return meta

  def RenderExtInfo(self, filename):
      return "#EXTINF: %d, %s" % (int(self.total_time),
                                  self.RenderFileInfo(filename))

  def __init__(self, fullpath, read_tags):

      base, ext = os.path.splitext(fullpath)
      ext = ext.lower()

      # defaults:
      self.total_time = 0
      self.samplerate = 'unkown'
      self.bitrate = 'unkown'
      self.mode = ''
      self.mode_extension = ''


      self.valid = 0
      self.artist = ''
      self.title = ''
      self.genre = ''
      self.comment = ''
      self.track = ''
      self.artist = ''
      self.album = ''
      self.year = ''

      try:
          info = MutagenInfo(fullpath)
          self.__dict__.update(info.__dict__)
      except Exception, e:
          print "Mutagen exception", str(e)

      #    if hasattr(info, 'length'):
      if self.total_time > 3600:
          self.duration = '%d:%02d:%02d' % (int(self.total_time / 3600),
                                            int(self.total_time / 60) % 60,
                                            int(self.total_time) % 60)
      elif self.total_time > 60:
          self.duration = '%d:%02d' % (int(self.total_time / 60),
                                       int(self.total_time) % 60)
      else:
          self.duration ='%02d' % int(self.total_time)

      self.display_name = self.RenderFileInfo(fullpath)

def ToUtf8(u):
    if  type(u) == unicode:
        return u.encode('utf8')

    assert type(u) == str
    a = ""
    try:
        a = u.decode('utf8')
    except Exception, e:
        #print "\nERROR-UTF8:", repr(u)
        pass

    if not a:
        try:
            a = u.decode('latin1')
        except Exception, e:
            print "\nERROR-LATIN1:", str(e)
            pass
    return a.encode('utf8')




class MutagenInfo:
  """Extra information about  pretty much any music file.
  http://www.sacredchao.net/quodlibet/wiki/Development/Mutagen
  """

  def KeyToStr(self, key):
      if type(key) not in [unicode, str]:
          try:
              key = str(key)
          except Exception, e:
              print "bad mutagen keys in: ", name, repr(key), str(e)
              return None
      key = key.lower()
      return key

  def ValToStr(self, val):
      if type(val) == list:
          val = val[0]
      if type(val) not in [unicode, str]:
          val = str(val)
      try:
          val = ToUtf8(val)
      except Exception, e:
          print "\nERROR-XXX:", str(e)
          return ""
      return val

  def __init__(self, name):
    #
    # Generic File Info
    #
    mf = mutagen.File(name)
    self.title = os.path.split(name)[1]

    if not mf:
      #print "WARNING - no tags for: ", name
      return

    # recognized_comments = ('Artist', 'Album', 'Title', 'Version',
    #                        'Organization', 'Genre', 'Description',
    #                        'Date', 'Location', 'Copyright', 'Vendor')
    try:
        self.total_time = mf.info.length
        self.samplerate = mf.info.sample_rate/1000
        filesize = os.stat(name).st_size
        try:
            self.bitrate = mf.info.bitrate/1000
        except:
          # an approximation since we do count meta info
            self.bitrate = int(filesize * 8 / self.total_time / 1000)
        self.filesize = filesize/1000000
    except:
        pass

    for key, val in mf.items():
        #print "@@@ [%s] [%s]" % (key, val)
        key = self.KeyToStr(key)

        # note that we do not compute the val unless we
        # match the key
        if key in ['tit2', 'title']:
            self.title = self.ValToStr(val)
        elif key in ['tpe1', 'artist']:
            self.artist = self.ValToStr(val)
        elif key in ['talb', 'album']:
            self.album = self.ValToStr(val)
        elif key in ['tcon', 'genre']:
            self.genre = self.ValToStr(val)
        elif key in ['tracknumber', 'track', 'trck']:
            self.track = self.ValToStr(val)
        elif key in ['year', 'tdrc', 'date']:
            self.year = self.ValToStr(val)
        elif key in ['comment'] or key.startswith("comm::"):
            self.comment = self.ValToStr(val)
        else:
            pass
            #print "@@@ UNKNOWN TAG", repr(key)
    self.valid = 1


def RenderExtInfo(filename, read_tags):
  if os.path.exists(filename):
      return FileInfo(filename, read_tags).RenderExtInfo(filename)
  else:
    return ""

def RenderIcyName(filename, read_tags):
  if os.path.exists(filename):
      return FileInfo(filename, read_tags).display_name
  else:
      return RenderFilename(filename)


if __name__ == '__main__':
    import sys
    for filename  in sys.argv[1:]:
        fi = FileInfo(filename, 1)
        print "title", fi.title
        print "artist", fi.artist
        print "album", fi.album
        print "track", fi.track
        print "year", fi.year
        print "comment", fi.comment
        print "genre", fi.genre

