#!/usr/bin/env python
# ======================================================================
# Copyright (C) 2008 and onwards Robert Muth <robert at muth dot org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ======================================================================

import cgi
import logging
import os
import re
import stat
import time
import urllib
import zipfile

try:
  import cStringIO
  StringIO = cStringIO
except ImportError:
  import StringIO

import converter
import helper
import fileinfo
import handler_stats
import mime_type
import playlist
import templates

SERVE_DIRS = []
SERVE_DICT = {}
ZIPSIZE_IN_FLIGHT = 0

######################################################################
##
######################################################################
class _datablob:
  def __init__(self, **args):
    self.__dict__.update(args)

######################################################################
##
######################################################################
class empty_delegator:
  "Delegate attrs to another object; fill in empty string for unknown attrs."
  def __init__(self, ob):
    self.ob = ob

  def __getattr__(self, name):
    if hasattr(self.ob, name) and getattr(self.ob, name):
      return getattr(self.ob, name)
    else:
      return "&nbsp;"

######################################################################
##
######################################################################
def ResetServeDirs():
    global SERVE_DIRS, SERVE_DICT
    SERVE_DIRS = []
    SERVE_DICT = {}

def AddServeDir(directory, nickname):
    global SERVE_DIRS, SERVE_DICT
    assert nickname not in SERVE_DICT
    assert "/" not in nickname
    if not os.path.isdir(directory):
        logging.error("directory [%s] does not exits", directory)
        logging.info(" skipping: dir %s (%s)" % (directory, tag))
    SERVE_DIRS.append(nickname)
    SERVE_DICT[nickname] = directory
    return


def HaveMultipleServeDirs():
    return len(SERVE_DIRS) > 1


def GetFirstServeDir():
    return SERVE_DICT[SERVE_DIRS[0]], SERVE_DIRS[0]


def NickNameToServerDir(nickname):
    if nickname not in SERVE_DICT:
        raise ValueError, "bad nickname %s" % nickname
    else:
        return SERVE_DICT[nickname]


def StripServerDir(path, nickname):
    dir = NickNameToServerDir(nickname)
    if not path.startswith(dir):
        logging.error("bad path [%s]", path)
        return ""
    return path[len(dir):]

######################################################################
##
######################################################################
def BuildUrl(tag, path, args={}):
#    host = (self.server.config.server.name_prefix or
#            self.headers.getheader('host') or
#            self.server.server_name)
#    if host.find(':') < 0:
#        host = "%s:%s" % (host, self.server.server_port)
    path = path.replace("\\", "/")  # if we're on Windows
    params = helper.MakeUrlParms(args)
#    return 'http://%s/%s%s%s' % (host, tag, helper.QuotePath(path), params)
    return '/%s%s%s' % (tag, helper.QuotePath(path), params)


def BuildDirUrl(dir, nickname):
    dir = StripServerDir(dir, nickname)
    url = BuildUrl("dir/" + nickname, dir)
    return url

def BuildPlaylistUrl(dir, nickname):
    dir = StripServerDir(dir, nickname)
    return BuildUrl("playlist/" + nickname, dir)


def BuildPseudoUrl(dir, nickname):
    dir = StripServerDir(dir, nickname)
    return BuildUrl("pseudo/" + nickname, dir)


def BuildZipUrl(dir, type, nickname):
    assert type in ["all.zip"]
    dir = os.path.join(dir, type)
    dir = StripServerDir(dir, nickname)
    return BuildUrl("zip/" + nickname, dir)


def BuildSingletonPlaylist(filename, nickname):
    filename = StripServerDir(filename, nickname)
    return BuildUrl("singleton/" + nickname , filename)


def BuildMediaUrl(filename, nickname, args={}):
    filename = StripServerDir(filename, nickname)
    return BuildUrl("media/" + nickname, filename, args)


######################################################################
##
######################################################################
def HandlerMedia(request, path, args, config, profile):
    nickname = path.pop(0)
    dir = NickNameToServerDir(nickname)
    filename = helper.PathJoin(dir, path)

    range = request.headers.getheader('range')
    ip, port = request.client_address
    handler_stats.LogUser(ip, time.time(), filename)

    if "cmedia" in args and args['cmedia'] != "none":
        # will block until job is done
       job = converter.SubmitJob(filename, args['cmedia'])
       job.WaitForCompletion()
       converter.PrefetchJob(filename, args['cmedia'])
       if job.IsFailure():
         request.send_error(403, 'Error converting file.')
         return
       filename = job.GetCovertedFile()
    type = mime_type.MimeTypeByExtension(filename)

    f = open(filename, 'rb')
    st = os.fstat(f.fileno())
    clen = st.st_size
    mtime = st.st_mtime
    read_tags = config.server.fileinfo
    icyname = None
    if mime_type.FileIsLikelySong(filename):
        icyname = "icy: " + fileinfo.RenderIcyName(filename, read_tags)
    helper.ServeRawData(request, f, type, clen, mtime, icyname, range)
    return

######################################################################
##
######################################################################
def HandlerZip(request, path, args, config, profile):
    global ZIPSIZE_IN_FLIGHT
    nickname = path.pop(0)
    dir = NickNameToServerDir(nickname)
    filename = helper.PathJoin(dir, path)
    dir, name = os.path.split(filename)
    assert name == 'all.zip'
    zipmax = config.server.max_zip_mb * 1024 * 1024
    if zipmax <= 0:
        request.send_error(403, 'The ZIP service has been disabled '
                        'by the server administrator.')
        return


    fout = StringIO.StringIO()
    z = zipfile.ZipFile(fout, 'w', zipfile.ZIP_STORED)
    songs = helper.MakeSongList(dir, 1)
    for s in songs:
        z.write(s, s)
        if ZIPSIZE_IN_FLIGHT + len(fout.getvalue()) > zipmax:
            break

    z.close()
    fout.seek(0)
    clen = len(fout.getvalue())
    logging.info("ZIP thresholds: %d + %d vs %d" %
                 (ZIPSIZE_IN_FLIGHT, clen, zipmax))

    if ZIPSIZE_IN_FLIGHT + clen > zipmax:
        request.send_error(503, 'The <b>ZIP</b> service is currently under '
                           'heavy load.  Please try again later.')
        return

    type = mime_type.MimeTypeByExtension(name)
    ZIPSIZE_IN_FLIGHT += clen
    range = request.headers.getheader('range')
    helper.ServeRawData(request, fout, type, clen, None, None, range)
    ZIPSIZE_IN_FLIGHT -= clen
    return


######################################################################
##
######################################################################
def HandlerDir(request, path, args, config, profile):
    nickname = path.pop(0)
    curdir = NickNameToServerDir(nickname)
    curdir = helper.PathJoin(curdir, path)

    albums = []
    pictures = []
    subdirs = []
    songs = []
    playlists = []
    plainfiles = []

    #
    pseudo = _datablob()
    pseudo.playlist = BuildPseudoUrl(curdir, nickname)

    # template currently crashes without this
    pseudo.zip = BuildZipUrl(curdir, "all.zip", nickname)

    title = cgi.escape(StripServerDir(curdir, nickname))
    # TODO: config use config
    read_tags = 1

    for name in helper.SortedDirectoryWithoutHiddenFiles(curdir):
        filename = os.path.join(curdir, name)
        try:
            age = _GetFileAge(filename)
        except Exception, e:
            # For example, in the case of disk I/O errors
            logging.error("Failed to stat %s - %s", name, str(e))
            continue
        if _FileShouldBeHidden(name, config):
            logging.info("Hiding %s", filename)
            continue

        base, ext = os.path.splitext(name)
        ext = ext.lower()

        if mime_type.FileIsLikelyPicture(name):
            blob = _datablob(href=BuildMediaUrl(filename, nickname),
                             age=age)
            pictures.append(blob)

        elif mime_type.FileIsLikelyPlaintext(name):
            blob = _datablob(href=BuildMediaUrl(filename, nickname),
                             age=age,
                             text=base)
            plainfiles.append(blob)

        elif mime_type.FileIsLikelyPlaylist(name):
            blob = _datablob(href=BuildPlaylistUrl(filename, nickname),
                             age=age,
                             text=base)
            playlists.append(blob)

        elif mime_type.FileIsLikelySong(name):
            # TODO: needs escapes
            info = empty_delegator(fileinfo.FileInfo(filename, read_tags))
            blob = _datablob(href=BuildMediaUrl(filename, nickname),
                             age=age,
                             filename=name,
                             info=info)
            blob.playlist = BuildSingletonPlaylist(filename, nickname)
            songs.append(blob)
            if not albums:
                blob = _datablob(query= urllib.quote(info.album + " " + info.artist))
                albums.append(blob)

        elif os.path.isdir(filename):
            playlist = BuildPseudoUrl(filename, nickname)

            blob = _datablob(href=BuildDirUrl(filename, nickname),
                             age=age,
                             playlist=playlist,
                             text=name)
            subdirs.append(blob)
        else:
            logging.error("do not know what to do with [%s]", filename)
    links = TreePosition(curdir, nickname)
    _DisplayPage(request, profile, title, subdirs, pictures, albums,
                 plainfiles, songs, playlists, pseudo, links)


def _DisplayPage(request, profile, title,
                 subdirs=[], pictures=[], albums=[], plainfiles=[],
                 songs=[], playlists=[], pseudo=None, links="", skiprec=0):

    template = templates.GetTemplate(profile["template"])
    content_type = 'text/html'

    request.send_response(200)
    request.send_header("Content-Type", content_type)
    request.end_headers()

    data = { 'title' : title,
             'links' : links,
             'pictures' : pictures,
             'albums' : albums,
             'plainfiles' : plainfiles,
             'subdirs' : subdirs,
             'songs' : songs,
             'pseudo': pseudo,
             'playlists' : playlists,
             }

    if not skiprec:
      data['display-recursive'] = 'yes'
    else:
      data['display-recursive'] = ''

    template.generate(request.wfile, data)

def TreePosition(curpath, nickname):
    links = []
    first = BuildDirUrl(NickNameToServerDir(nickname), nickname)

    last = 1
    while 1:
      url = BuildDirUrl(curpath, nickname)
      if url == first:
        break
      curpath, name = os.path.split(curpath)
      name = cgi.escape(name)
      if last:
        links.append('<b> / %s</b>' % name)
      else:
        links.append('<b> / </b><a href="%s/">%s</a>' % (url, name))
      last = 0

    if HaveMultipleServeDirs():
      links.append('<b> / </b><a href="%s/">%s</a>\n' % (first, nickname))
      links.append('<a href="/">HOME</a>\n')
    else:
      links.append('<a href="%s/">HOME</a>\n' % first)
    links.reverse()
    return '<p>' + '\n'.join(links) + '</p>'


def HandlerSingleton(request, path, args, config, profile):
      nickname = path.pop(0)
      dir = NickNameToServerDir(nickname)
      filename = helper.PathJoin(dir, path)
      ServePlaylistGeneric(request, [filename], nickname, args)
      return


def HandlerPseudo(request, path, args, config, profile):
      nickname = path.pop(0)
      dir = NickNameToServerDir(nickname)
      dir = helper.PathJoin(dir, path)

      songs = helper.MakeSongList(dir, args.has_key("recursive"))
      if args.has_key("shuffle"): helper.ShuffleList(songs)
      ServePlaylistGeneric(request, songs, nickname, args)
      return


def HandlerPlaylist(request, path, args, config, profile):
      nickname = path.pop(0)
      dir = NickNameToServerDir(nickname)
      filename = helper.PathJoin(dir, path)
      mtime = os.stat(filename)[stat.ST_MTIME]
      pl = playlist.Playlist(0, dir)
      data = open(filename)
      if filename.endswith(".m3u"):
          parser = playlist.M3uParser()
          parser.Parse(data, pl)
      elif filename.endswith(".xspf"):
          parser = playlist.XspfParser()
          parser.Parse(data, pl)
      else:
          print "unsupported playlist format: %s" % filename
          assert 0

      songs = pl.track[:]
      ServePlaylistGeneric(request, songs, nickname, args)
      return


def ServePlaylistGeneric(request, songs, nickname, args):
      extinfo = args.has_key("extinfo")
      ext = args.get("cplaylist", ".xspf")
      pl = playlist.Playlist()
      args_media = {}
      k = "cmedia"
      if k in args:
        args_media[k] = args[k]
      last_song = None
      for s in songs:
          meta = {}
          if extinfo:
              meta = fileinfo.FileInfo(s.strip(), 1).GetDict()
          song_url = BuildMediaUrl(s, nickname, args_media)

          pl.AddTrack(song_url, meta)
          if last_song:
            converter.AddPrefetch(last_song, s)
          last_song = s
      if ext == ".m3u":
          fout = StringIO.StringIO(pl.MakeM3u(extinfo, 0))
      elif ext == ".xspf":
          fout = StringIO.StringIO(pl.MakeXspf(extinfo))
      else:
          print "unsupported playlist format %s" % ext
          assert 0
      clen = len(fout.getvalue())
      type = mime_type.MimeTypeByExtension("dummy" + ext)

      helper.ServeRawData(request, fout, type, clen)
      return

def _FileShouldBeHidden(name, config):
      for regexp in config.directory.hide_names:
          if re.search(regexp, name):
              return 1
      return 0

  # return empty string or a "new since..." string
def _GetFileAge(filename):
    return os.stat(filename)[stat.ST_MTIME]


######################################################################
##
######################################################################
def HandlerToplevel(request, path, args, config, profile):
    """This is only used when there are multile top level dirs"""

    if not HaveMultipleServeDirs():
      dir, nickname = GetFirstServeDir()
      url = BuildDirUrl(dir, nickname)
      request.redirect(url)
      return

    # TODO: move this to handler_media
    subdirs = []

    # TODO: do not use config here
    for nickname, filename in config.media.dirs:
        blob = _datablob(href=BuildDirUrl(filename, nickname),
                         age=0,
                         playlist='',
                         text=nickname)
        subdirs.append(blob)
    _DisplayPage(request, profile, config.media.title,
                 subdirs=subdirs, skiprec=1)

    return

######################################################################
##
######################################################################
def HandlerResources(request, path, args, config, resources):
    if len(path) != 1:
      logging.error("code 404 -- bad resource")
      request.send_response(404)
      request.send_header("Content-Type", "text/plain")
      request.end_headers()
      return
    filename = helper.PathJoin(config.server.resource_path, path)
    type = mime_type.MimeTypeByExtension(filename)
    f = open(filename, 'rb')
    st = os.fstat(f.fileno())
    clen = st.st_size
    mtime = st.st_mtime
    helper.ServeRawData(request, f, type, clen, mtime)
    return

