#!/usr/bin/env python
# ======================================================================
# Copyright (C) 2008 and onwards Robert Muth <robert at muth dot org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ======================================================================

import logging
import os
import random
import socket
import time
import urllib


import mime_type
######################################################################
##
######################################################################
class ClientAbortedException(Exception):
    pass

######################################################################
##
######################################################################
def SortedDirectoryWithoutHiddenFiles(d):
    l = [f for f in  os.listdir(d) if f[0] != '.']
    l.sort()
    return l

def dot2int(dotaddr):
    a, b, c, d = map(int, dotaddr.split('.'))
    return (a << 24) + (b << 16) + (c << 8) + (d << 0)

def EscapeContents(contents):
    # TODO: find better library function for this
    contents = contents.replace("&", "&amp;")
    contents = contents.replace("<", "&lt;")
    contents = contents.replace(">", "&gt;")
    contents = contents.replace("[", " ")
    contents = contents.replace("]", " ")
    return contents

def QuotePath(path):
    return urllib.quote(urllib.quote(path))

def UnquotePath(path):
    return urllib.unquote(urllib.unquote(path))

def PathJoin(prefix, path):
    for p in path:
        prefix = os.path.join(prefix, p)
    return prefix

def MakeUrlParms(args):
    if not args:
        return ""
    params = ""
    prefix = "?"
    for k in args:
        params += prefix + k
        if args[k] is not None:
            params += "=" + args[k]
        prefix = "&"
    return params

def ShuffleList(songs):
    count = len(songs)
    for i in xrange(count):
      j = random.randrange(count)
      songs[i], songs[j] = songs[j], songs[i]

def MakeSongList(fullpath, recursive):
    """
    This routine takes a string for 'fullpath', and boolean for 'recursive'.
    If recursive is false make_list will return a list of every file ending
    in '.mp3' in fullpath.
    If recursive is true make_list will return a list of every
    file ending in '.mp3' in fullpath and in every directory beneath
    fullpath.

    WARNING: There is no checking for the recursive directory structures
    which are possible in most Unixes using ln -s etc...  If you have
    such a directory structure, make_list will continue to traverse it
    until it hits the inherent limit in Python for the number of functions.
    This number is quite large. I found this out the hard way :). Learn
    from my experience...
    """
    songs = []

    for name in SortedDirectoryWithoutHiddenFiles(fullpath):
      filename = os.path.join(fullpath, name)
      if os.path.isfile(filename):
        if mime_type.FileIsLikelySong(filename):
          # add the song's URL to the list we're building
          songs.append(filename)
      elif os.path.isdir(filename):
        if recursive:
          songs += MakeSongList(filename, recursive)
    return songs


def ServeRawData(request, f, type, clen,
                 mtime=None, icyname=None, range=None):
    """
    Serve raw data, including range requests
    """
    if range:
        request.send_response(206)
    else:
        request.send_response(200)
    request.send_header("Content-Type", type)

    if mtime:
      request.send_header('Last-Modified', time.strftime("%a, %d %b %Y %T GMT"))
    if icyname:
      request.send_header("icy-name", icyname.replace(" ","_"))
    #print request.headers
    if range:
        #print "@@@@ range given", range
        type, seek = range.split('=')
        startSeek, endSeek = seek.split('-')
        startSeek = int(startSeek)
        if endSeek:
            endSeek = int(endSeek)
        else:
            pass
            endSeek = clen - 1
            range = range + str(clen)
        f.seek(startSeek)
        request.send_header("Content-Range",
                            "bytes %d-%d/%d" % (startSeek, endSeek, clen))
        clen = 1 + endSeek - startSeek
        #print "@@@@ range out", startSeek, endSeek, clen

    request.send_header("Content-Length", clen)
    request.end_headers()


    while clen > 0:
      data = f.read(min(8192, clen))
      if not data:
        logging.error("file longer than expected")
        break
      clen -= len(data)
      try:
        request.wfile.write(data)
      except ClientAbortedException:
        logging.error('client closed connection for [%s]', request.path)
        break
      except socket.error:
        logging.error('socket error [%s]', request.path)
        # it was probably closed on the other end
        break
