#!/usr/bin/env python
# ======================================================================
# Copyright (C) 2008 at onward Robert Muth <robert @ muth dot org>.
# All Rights Reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ======================================================================

""" frieda playlist module

for a description of the m3u format see
http://hanna.pyxidis.org/tech/m3u.html

for a description of the xspf format see
http://xspf.org/quickstart/
"""

# python imports
import logging
import urllib
import xml.sax
import xml.sax.saxutils
import os

EXTM3U = '#EXTM3U'
EXTINF = '#EXTINF:'
FILE_URL_PREFIX = "file://"
DURATION = "duration"
TITLE = "title"

XSPF_TITLE = ["playlist" , "title"]
XSPF_CREATOR = ["playlist" , "creator"]
XSPF_INFO = ["playlist" , "info"]

XSPF_TRACK = ["playlist", "trackList", "track"]
XSPF_TRACK_LOCATION = XSPF_TRACK + ["location"]
XSPF_TRACK_DURATION = XSPF_TRACK + ["duration"]
XSPF_TRACK_CREATOR = XSPF_TRACK + ["creator"]
XSPF_TRACK_TITLE = XSPF_TRACK + ["title"]


def list_eq(l1, l2):
    if len(l1) != len(l2): return 0
    for e1, e2 in zip(l1, l2):
        if e1 != e2: return 0
    return 1

def dict_eq(d1, d2):
    if len(d1) != len(d2): return 0
    for k in d1:
        if k not in d2 or d1[k] != d2[k]: return 0
    return 1


class XspfParser(xml.sax.handler.ContentHandler):
    def __init__(self):
        self._path = []
        self._content = ""
        self._playlist = None
        self._location = ""
        self._meta = {}
        return

    def Parse(self, s, playlist):
        self._path = []
        self._content = ""
        self._playlist = playlist
        try:
            xml.sax.parseString(s, self)
        except xml.sax.SAXParseException:
          logging.warning("bad xspf file")

    def startElement(self, name, attrs):
        self._path.append(name)
        self._content = ""
        if list_eq(self._path, XSPF_TRACK):
            self._location = ""
            self._meta = {}

    def characters(self, content):
        self._content += xml.sax.saxutils.unescape(content)

    def endElement(self, name):
        if list_eq(self._path, XSPF_TITLE):
            self._playlist.title = xml.sax.saxutils.unescape(self._content)
        elif list_eq(self._path, XSPF_CREATOR):
            self._playlist.creator = xml.sax.saxutils.unescape(self._content)
        elif list_eq(self._path, XSPF_INFO):
            self._playlist.info = xml.sax.saxutils.unescape(self._content)
        elif list_eq(self._path, XSPF_TRACK):
            self._playlist.AddTrack(self._location, self._meta)
        elif list_eq(self._path, XSPF_TRACK_LOCATION):
            self._location = self._content.strip()
        elif list_eq(self._path, XSPF_TRACK_DURATION):
            self._meta[DURATION] = xml.sax.saxutils.unescape(self._content.strip())
        elif list_eq(self._path, XSPF_TRACK_TITLE):
            self._meta[TITLE] =  xml.sax.saxutils.unescape(self._content.strip())


        assert self._path
        self._path.pop(-1)

class M3uParser:
    def __init__(self):
        return

    def Parse(self, s, playlist):
        meta = {}
        location = ""
        for line in s:
            line = line.strip()
            if not line:
                continue
            if line.startswith("#"):
                if line.startswith(EXTM3U):
                    pass
                elif line.startswith(EXTINF) and "," in line:
                    durartion, title = line[len(EXTINF):].split(",", 2)
                    meta[DURATION] = durartion
                    meta[TITLE] = title
                else:
                    logging.warning("unknown comment in m3u file: %s", line)
                continue
            else:
                location = line
            playlist.AddTrack(location, meta)
            meta = {}
            location = ""
        return


class Playlist:

    def __init__(self, translate_path=0, base=""):
        self.title = ""
        self.creator = ""
        self.info = ""

        self.track = []
        self.meta = []
        self._translate_path = translate_path
        self._base = base
        return

    def TranslateLocation(self, location):
        if (location.startswith("http:") or
            location.startswith("ftp:") or
            location.startswith("file:")):
            return location
        else:
          if self._base and not os.path.isabs(location):
            location = os.path.join(self._base, location)
          if self._translate_path:
            location = FILE_URL_PREFIX + urllib.pathname2url(location)
          return location

    def IsEqual(self, pl):
      if self.title != pl.title: return 0
      if self.creator != pl.creator: return 0
      if self.info != pl.info: return 0
      if not list_eq(self.track, pl.track): return 0
      if len(self.meta) != len(pl.meta): return 0
      for d1, d2 in zip(self.meta, pl.meta):
          if not dict_eq(d1, d2): return 0
      return 1

    def AddTrack(self, location, meta):
        assert len(location) > 1
        location = self.TranslateLocation(location)
        self.track.append(location)
        self.meta.append(meta)
        return

    def MakeM3u(self, extm3u=1, generate_file_prefix=1):
        assert len(self.meta) == len(self.track)
        s = []
        if extm3u:
            s.append(EXTM3U + "\n")
        for meta, track in zip(self.meta, self.track):
            if extm3u and DURATION in meta and TITLE in meta:
                s.append(EXTINF + meta[DURATION] + "," + meta[TITLE] + "\n")
            if not generate_file_prefix and track.startswith(FILE_URL_PREFIX):
              track = urllib.url2pathname(track[len(FILE_URL_PREFIX):])
            s.append(track + "\n")
        return "".join(s)


    def MakeXspf(self, annotate=1):
        s = []
        s.append("<?xml version='1.0' encoding='UTF-8'?>\n")
        s.append("<playlist version='1'>\n")
        s.append("<title>%s</title>\n" % xml.sax.saxutils.escape(self.title))
        s.append("<creator>%s</creator>\n" % xml.sax.saxutils.escape(self.creator))
        s.append("<info>%s</info>\n" % xml.sax.saxutils.escape(self.info))
        s.append("<trackList>\n")
        for meta, track in zip(self.meta, self.track):
            s.append("    <track>\n")
            if annotate and meta:
                for key, value in meta.items():
                    if key == "artist": key = "creator"
                    s.append("    <%s>%s</%s>\n" %
                             (key, xml.sax.saxutils.escape(value), key))
            s.append("    <location>%s</location>\n" % track)
            s.append("    </track>\n")
        s.append("</trackList>\n")
        s.append("</playlist>\n")
        return "".join(s)

