#!/usr/bin/env python
# ======================================================================
# Copyright (C) 2008 at onward Robert Muth <robert @ muth dot org>.
# All Rights Reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ======================================================================


"""request handler module for most requests"""
# python imports
import SocketServer
import BaseHTTPServer
import logging
import os
import re
import socket

try:
  import cStringIO
  StringIO = cStringIO
except ImportError:
  import StringIO

# local imports
import authentication
import config_helper
import helper
import handler_media
import handler_stats
import templates


######################################################################
##
######################################################################
class RequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

  def do_GET(self):
      try:
          self._perform_GET()
      except helper.ClientAbortedException:
          logging.info('ClientAbortException')
      except Exception, e:
          logging.error("toplevel exception %s", repr(e))
          raise

  def SendAuthenticationRequest(self):
    realm='frieda'
    self.send_response(401)
    self.send_header('WWW-Authenticate', 'Basic realm="%s"' % realm)
    self.send_header('Content-Type', 'text/html;');
    self.send_header('Connection', 'close');
    self.end_headers()
    try:
      short, long = self.responses[401]
    except KeyError:
      short, long = '???', '???'

    self.wfile.write(self.error_message_format %
                     {'code': 401, 'message': short, 'explain': long})


  def _perform_GET(self):
    Handlers = {
      'top':  handler_media.HandlerToplevel,

      'stats':  handler_stats.HandlerStats,
      'logout': HandlerLogout,
      'quitquitquit': HandlerQuit,
      'config': config_helper.HandlerConfig,
      'resources': handler_media.HandlerResources,
      'media': handler_media.HandlerMedia,
      'dir': handler_media.HandlerDir,
      'playlist': handler_media.HandlerPlaylist,
      'pseudo': handler_media.HandlerPseudo,
      'singleton': handler_media.HandlerSingleton,
      'zip': handler_media.HandlerZip,
      }

    path, args = self.translate_path()
    if path is None:
      self.send_error(400, 'Illegal URL construction')
      return


    if not path:
      path = ["top"]
    kind = path.pop(0)
    if kind in ['favicon.ico', 'robots.txt']:
      path = [kind]
      kind = 'resources'

    user, profile = authentication.GetProfile(self,
                                              self.server.config.acl,
                                              [kind] + path)
    if not profile:
      return

    handler_stats.IncCounter("user", user)

    config = self.server.config

    if kind in Handlers:
      handler_stats.IncCounter("dispatch", kind)
      Handlers[kind](self, path, args, config, profile)
    else:
      self.send_error(404)
    return

  def translate_path(self):
    args = {}
    if "?" in self.path:
      self.path, params = self.path.split("?", 2)
      params = params.split("&")
      for p in params:
        if "=" in p:
          k, v= p.split("=", 2)
          args[k] = v
        else:
          args[p] = None

    parts = helper.UnquotePath(self.path).split('/')
    parts = filter(None, parts)
    while 1:
      try:
        parts.remove('.')
      except ValueError:
        break
    while 1:
      try:
        idx = parts.index('..')
      except ValueError:
        break
      if idx == 0:
        logging.info('Warning in translate_path --- Illegal path:'
                     ' the \'..\' attempted to go above the root')
        return None
      del parts[idx-1:idx+1]
    return parts, args

  def redirect(self, url):
    "Send a redirect to the specified URL."
    logging.error("code 301 -- Moved")
    self.send_response(301, 'Moved')
    self.send_header('Location', url)
    self.end_headers()
    self.wfile.write(self.error_message_format %
                     {'code': 301,
                      'message': 'Moved',
                      'explain': 'Object moved permanently'})

  def setup(self):
    SocketServer.StreamRequestHandler.setup(self)

    # wrap the wfile with a class that will eat up "Broken pipe" errors
    self.wfile = _SocketWriter(self.wfile)

  def finish(self):
    # if the other end breaks the connection, these operations will fail
    try:
      self.wfile.close()
    except socket.error:
      pass
    try:
      self.rfile.close()
    except socket.error:
      pass

######################################################################
##
######################################################################
def HandlerLogout(request, path, args, config, setting):
    request.SendAuthenticationRequest()
    # request.send_response(401)
    return

######################################################################
##
######################################################################
def HandlerQuit(request, path, args, config, setting):
    request.server.StopServer()
    request.send_response(404)
    return

######################################################################
##
######################################################################
class _SocketWriter:
  "This class ignores 'Broken pipe' errors."
  def __init__(self, wfile):
    self.wfile = wfile

  def __getattr__(self, name):
    return getattr(self.wfile, name)

  def write(self, buf):
    try:
      s_buf = str(buf)
      return self.wfile.write(s_buf)
    except IOError, v:
      if v.errno in [32, 104]:
        raise helper.ClientAbortedException
      # not a 'Broken pipe' or Connection reset by peer
      # re-raise the error
      raise
    except socket.error, v:
      if v[0] in [32, 104]:
        raise helper.ClientAbortedException
      raise

